import numpy as np
import pandas as pd


def preprocessing_esg_ratings(data: pd.DataFrame):
    # Define a mapping from ratings to numerical scores
    rating_map = {"A+": 6, "A": 5, "B+": 4, "B": 3, "C": 2, "D": 1, np.nan: np.nan}

    # Convert rating columns into numerical scores
    for year in range(2011, 2021):
        data[str(year)] = data[str(year)].map(rating_map)

    # Melt dataframe to have years as rows instead of columns
    melted_df = data.melt(
        id_vars=["type", "code", "name", "market"],
        var_name="year",
        value_name="rating_score",
        value_vars=[str(year) for year in range(2011, 2021)],
    )

    # Convert year to integer
    melted_df["year"] = melted_df["year"].astype(int)

    # Sort by type, code and year
    melted_df.sort_values(["type", "code", "year"], inplace=True)

    # Calculate rating change
    melted_df["rating_change"] = melted_df.groupby(["type", "code"])[
        "rating_score"
    ].diff()

    # Create 'rating' column by mapping numerical scores back to ratings
    reverse_rating_map = {v: k for k, v in rating_map.items()}
    melted_df["rating"] = melted_df["rating_score"].map(reverse_rating_map)

    # Set rating_change values to '1', '0', '-1' based on whether it's positive, zero or negative
    melted_df["rating_change"] = melted_df["rating_change"].apply(
        lambda x: 1 if x > 0 else (0 if x == 0 else (-1 if x < 0 else np.nan))
    )
    return melted_df



"""
Type aliases for datasets.
"""
from typing import Union

from datasets.arrow_dataset import Dataset
from datasets.dataset_dict import DatasetDict, IterableDatasetDict
from datasets.iterable_dataset import IterableDataset

DatasetType = Union[Dataset, IterableDataset]
DatasetDictType = Union[DatasetDict, IterableDatasetDict]
DatasetLikeType = Union[
    Dataset,
    IterableDataset,
    DatasetDict,
    IterableDatasetDict,
]



"""
Graphics functions
"""


class GRAPHICs:
    ###############################
    # Graphics functions
    ###############################
    @staticmethod
    def collage(
        images_or_uris,
        collage_filepath=None,
        ncols=3,
        max_images=12,
        collage_width=1200,
        padding: int = 10,
        bg_color: str = "black",
        crop_to_min_size=False,
        show_filename=False,
        filename_offset=(5, 5),
        fontname=None,
        fontsize=12,
        fontcolor="#000",
        **kwargs,
    ):
        from ._collage import collage as _collage

        return _collage(
            images_or_uris,
            collage_filepath=collage_filepath,
            ncols=ncols,
            max_images=max_images,
            collage_width=collage_width,
            padding=padding,
            bg_color=bg_color,
            crop_to_min_size=crop_to_min_size,
            show_filename=show_filename,
            filename_offset=filename_offset,
            fontname=fontname,
            fontsize=fontsize,
            fontcolor=fontcolor,
            **kwargs,
        )

    @staticmethod
    def make_gif(
        image_filepaths=None,
        filename_patterns: str = "",
        base_dir: str = "",
        output_filepath: str = "",
        duration: int = 100,
        loop: int = 0,
        width: int = 0,
        optimize: bool = True,
        quality: int = 50,
        show: bool = False,
        force: bool = False,
        **kwargs,
    ):
        from hyfi.graphics.motion import make_gif as _make_gif

        return _make_gif(
            image_filepaths=image_filepaths,
            filename_patterns=filename_patterns,
            base_dir=base_dir,
            output_filepath=output_filepath,
            duration=duration,
            loop=loop,
            width=width,
            optimize=optimize,
            quality=quality,
            show=show,
            force=force,
            **kwargs,
        )

    @staticmethod
    def get_image_font(fontname: str = "", fontsize: int = 12):
        from .utils import get_image_font

        return get_image_font(fontname, fontsize)

    @staticmethod
    def load_image(
        image_or_uri,
        max_width: int = 0,
        max_height: int = 0,
        max_pixels: int = 0,
        scale: float = 1.0,
        resize_to_multiple_of: int = 0,
        crop_box=None,
        mode="RGB",
        **kwargs,
    ):
        from .utils import load_image

        return load_image(
            image_or_uri,
            max_width,
            max_height,
            max_pixels,
            scale,
            resize_to_multiple_of,
            crop_box,
            mode,
            **kwargs,
        )

    @staticmethod
    def scale_image(
        image,
        max_width: int = 0,
        max_height: int = 0,
        max_pixels: int = 0,
        scale: float = 1.0,
        resize_to_multiple_of: int = 8,
        resample: int = 1,
    ):
        """
        Scale an image to a maximum width, height, or number of pixels.

        resample:   Image.NEAREST (0), Image.LANCZOS (1), Image.BILINEAR (2),
                    Image.BICUBIC (3), Image.BOX (4) or Image.HAMMING (5)
        """
        from hyfi.graphics.utils import scale_image

        return scale_image(
            image,
            max_width,
            max_height,
            max_pixels,
            scale,
            resize_to_multiple_of,
            resample,
        )



"""
A class to run a pipeline.
"""
from typing import Any, Dict, Optional, Union

from hyfi.task import TaskConfig
from hyfi.utils.logging import LOGGING
from hyfi.workflow import WorkflowConfig

from .config import PipeConfig, PipelineConfig, run_pipe

logger = LOGGING.getLogger(__name__)


class PIPELINEs:
    """
    A class to run a pipeline.
    """

    @staticmethod
    def run_pipeline(
        config: Union[Dict, PipelineConfig],
        initial_object: Optional[Any] = None,
        task: Optional[TaskConfig] = None,
    ) -> Any:
        """
        Run a pipeline given a config

        Args:
            config: PipelineConfig to run the pipeline
            initial_obj: Object to use as initial value
            task: TaskConfig to use as task

        Returns:
            The result of the pipeline
        """
        if task is None:
            task = TaskConfig()
        return task.run_pipeline(config, initial_object)

    @staticmethod
    def run_pipe(
        obj: Any,
        config: Union[Dict, PipeConfig],
    ) -> Any:
        """
        Run a pipe on an object

        Args:
            obj: The object to pipe on
            config: The configuration for the pipe

        Returns:
            The result of the pipe
        """
        return run_pipe(obj, config)

    @staticmethod
    def run_task(
        task: TaskConfig,
        dryrun: bool = False,
    ):
        """
        Run pipelines specified in the task

        Args:
            task: TaskConfig to run pipelines for
            project: ProjectConfig to run pipelines
        """
        if dryrun:
            print("\nDryrun is enabled, not running the HyFI task\n")
            return
        task.run()

    @staticmethod
    def run_workflow(workflow: WorkflowConfig, dryrun: bool = False):
        """
        Run the tasks specified in the workflow

        Args:
            workflow: WorkflowConfig object to run
        """
        if dryrun:
            print("\nDryrun is enabled, not running the HyFI workflow\n")
            return
        workflow.run()

    @staticmethod
    def pipe(**kwargs) -> PipeConfig:
        """
        Return the PipeConfig.

        Args:
            **kwargs: Additional keyword arguments to pass to the PipeConfig constructor.

        Returns:
            PipeConfig: An instance of the PipeConfig class.
        """
        return PipeConfig(**kwargs)



from typing import Any, Dict, List, Optional, Union

from hyfi.composer import BaseModel, Composer, model_validator
from hyfi.pipeline.config import (
    PipelineConfig,
    Pipelines,
    RunningConfig,
    RunningTasks,
    get_running_configs,
)
from hyfi.project import ProjectConfig
from hyfi.task import TaskConfig
from hyfi.utils.contexts import elapsed_timer
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)

Tasks = List[Any]


class WorkflowConfig(BaseModel):
    _config_group_: str = "/workflow"
    _config_name_: str = "__init__"
    _auto_populate_: bool = True

    workflow_name: str = _config_name_
    project: Optional[ProjectConfig] = None
    task: Optional[TaskConfig] = None
    tasks: Optional[List[Union[str, Dict]]] = []
    pipelines: Optional[List[Union[str, Dict]]] = []
    verbose: bool = False

    @model_validator(mode="before")
    def validate_model_config_before(cls, data):
        # logger.debug("Validating model config before validating each field.")
        return Composer.to_dict(data)

    def get_running_tasks(self) -> RunningTasks:
        return get_running_configs(self.tasks or [])

    def get_running_task(self, rc: RunningConfig) -> Any:
        config = getattr(self, rc.uses, None)
        if rc.uses and isinstance(config, dict):
            if Composer.is_instantiatable(config):
                task = Composer.instantiate(config)
                if task is not None and getattr(task, "__call__", None):
                    return task
            else:
                task = TaskConfig(**config)
                task.name = rc.uses
                return task
        return None

    def get_task(self):
        return self.task or TaskConfig()

    def run(self):
        """
        Run the tasks specified in the workflow
        """
        if self.verbose:
            logger.info("Running %s task(s)", len(self.tasks or []))
        # Run all tasks in the workflow.
        with elapsed_timer(format_time=True) as elapsed:
            for rc in self.get_running_tasks():
                task = self.get_running_task(rc)
                task_name = (
                    task.task_name
                    if isinstance(task, TaskConfig)
                    else getattr(task, "_config_name_", "unknown")
                )
                logger.info("Running task [%s] with [%s]", task_name, rc)
                if isinstance(task, TaskConfig):
                    # Run the task if verbose is true.
                    task.run()
                elif task is not None and getattr(task, "__call__", None):
                    if rc.run_kwargs:
                        task(**rc.run_kwargs)
                    else:
                        task()
                else:
                    logger.warning("Invalid task: %s", task)
            # Print the elapsed time.
            if self.verbose:
                logger.info(
                    " >> elapsed time for the workflow with %s tasks: %s",
                    len(self.tasks or []),
                    elapsed(),
                )
        # Run the pipelines in the workflow, if any.
        if self.pipelines:
            task = self.get_task()
            logger.info(
                "Running pipelines in the workflow with task [%s]", task.task_name
            )
            task.run(
                pipelines=self.get_pipelines(),
            )

    def get_pipelines(self) -> Pipelines:
        """
        Get the list of pipelines for a workflow

        Args:
            task: The task to get the pipelines for

        Returns:
            A list of PipelineConfig objects
        """
        self.pipelines = self.pipelines or []
        pipelines: Pipelines = []
        for name in self.pipelines:
            if isinstance(name, str) and isinstance(getattr(self, name), dict):
                pipeline = PipelineConfig(**getattr(self, name))
                if not pipeline.name:
                    pipeline.name = name
                pipelines.append(pipeline)
        return pipelines



"""
Dataset transformation functions. Concatenate, merge, join, etc.
"""
from typing import Dict, List, Optional, Sequence, Tuple, Union

import datasets as hfds
import pandas as pd
from datasets.arrow_dataset import Dataset
from datasets.info import DatasetInfo
from datasets.splits import NamedSplit

from hyfi.utils.logging import LOGGING

from .types import DatasetType

logger = LOGGING.getLogger(__name__)


class DSCombine:
    @staticmethod
    def concatenate_data(
        data: Union[Dict[str, pd.DataFrame], Sequence[pd.DataFrame], List[DatasetType]],
        columns: Optional[Sequence[str]] = None,
        add_split_key_column: bool = False,
        added_column_name: str = "_name_",
        ignore_index: bool = True,
        axis: int = 0,
        split: Optional[str] = None,
        verbose: bool = False,
        **kwargs,
    ) -> Union[pd.DataFrame, DatasetType]:
        # if data is a list of datasets, concatenate them
        if isinstance(data, list) and isinstance(data[0], Dataset):
            return DSCombine.concatenate_datasets(
                data,
                axis=axis,
                split=split,
                **kwargs,
            )
        else:
            return DSCombine.concatenate_dataframes(
                data,
                columns=columns,
                add_split_key_column=add_split_key_column,
                added_column_name=added_column_name,
                ignore_index=ignore_index,
                axis=axis,
                verbose=verbose,
                **kwargs,
            )

    @staticmethod
    def concatenate_dataframes(
        data: Union[Dict[str, pd.DataFrame], Sequence[pd.DataFrame]],
        columns: Optional[Sequence[str]] = None,
        add_split_key_column: bool = False,
        added_column_name: str = "_name_",
        ignore_index: bool = True,
        axis: int = 0,
        verbose: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """Concatenate dataframes"""
        if isinstance(data, dict):
            dfs = []
            for split in data:
                df = data[split]
                if add_split_key_column:
                    df[added_column_name] = split
                dfs.append(df)
            data = dfs
        if add_split_key_column:
            columns.append(added_column_name)
        if isinstance(data, list):
            if isinstance(columns, list):
                _columns = [c for c in columns if c in df.columns]
                data = [df[_columns] for df in data]
            if verbose:
                logger.info("Concatenating %s dataframes", len(data))
            if len(data) > 0:
                return pd.concat(data, ignore_index=ignore_index, axis=axis)
            else:
                raise ValueError("No dataframes to concatenate")
        else:
            logger.warning("Warning: data is not a dict")
            return data

    @staticmethod
    def concatenate_datasets(
        dsets: List[DatasetType],
        info: Optional[DatasetInfo] = None,
        split: Optional[NamedSplit] = None,
        axis: int = 0,
    ) -> DatasetType:
        """
        Converts a list of [`Dataset`] with the same schema into a single [`Dataset`].

        Args:
            dsets (`List[datasets.Dataset]`):
                List of Datasets to concatenate.
            info (`DatasetInfo`, *optional*):
                Dataset information, like description, citation, etc.
            split (`NamedSplit`, *optional*):
                Name of the dataset split.
            axis (`{0, 1}`, defaults to `0`):
                Axis to concatenate over, where `0` means over rows (vertically) and `1` means over columns
                (horizontally).

                <Added version="1.6.0"/>

        Example:

        ```py
        >>> ds3 = concatenate_datasets([ds1, ds2])
        ```
        """
        return hfds.concatenate_datasets(
            dsets=dsets,
            info=info,
            split=split,
            axis=axis,
        )

    @staticmethod
    def merge_dataframes(
        left: pd.DataFrame,
        right: pd.DataFrame,
        how: str = "inner",
        on: Optional[Union[str, List[str]]] = None,
        left_on: Optional[Union[str, List[str]]] = None,
        right_on: Optional[Union[str, List[str]]] = None,
        left_index: bool = False,
        right_index: bool = False,
        sort: bool = False,
        suffixes: Tuple[str, str] = ("_x", "_y"),
        copy: bool = True,
        indicator: bool = False,
        validate: Optional[str] = None,
        verbose: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """Merge two dataframes

        Args:
            left (pd.DataFrame): left dataframe
            right (pd.DataFrame): right dataframe
            how (str, optional): how to merge. Defaults to "inner".
            on (Optional[Union[str, List[str]]], optional): column(s) to merge on. Defaults to None.
            left_on (Optional[Union[str, List[str]]], optional): column(s) to merge on in left dataframe. Defaults to None.
            right_on (Optional[Union[str, List[str]]], optional): column(s) to merge on in right dataframe. Defaults to None.
            left_index (bool, optional): merge on index in left dataframe. Defaults to False.
            right_index (bool, optional): merge on index in right dataframe. Defaults to False.
            sort (bool, optional): sort merged dataframe. Defaults to False.
            suffixes (Optional[Sequence[str]], optional): suffixes to add to column names. Defaults to None.
            copy (bool, optional): copy dataframes. Defaults to True.
            indicator (bool, optional): add indicator column. Defaults to False.
            validate (Optional[str], optional): validation method. Defaults to None.
            verbose (bool, optional): verbose logging. Defaults to False.

        Returns:
            pd.DataFrame: merged dataframe

        Example:
            >>> df1 = pd.DataFrame({"id": [1, 2, 3], "a": [1, 2, 3]})
            >>> df2 = pd.DataFrame({"id": [1, 2, 4], "b": [4, 5, 6]})
            >>> df = DSCombine.merge_dataframes(df1, df2, on="id")
            >>> df
                id  a  b
            0   1  1  4
            1   2  2  5

        """
        if verbose:
            logger.info("Merging dataframes")
        return left.merge(
            right,
            how=how,
            on=on,
            left_on=left_on,
            right_on=right_on,
            left_index=left_index,
            right_index=right_index,
            sort=sort,
            suffixes=suffixes,
            copy=copy,
            indicator=indicator,
            validate=validate,
            **kwargs,
        )



from ._collage import Collage, collage, gallery
from .graphics import GRAPHICs
from .motion import create_video, extract_frames, make_gif
from .utils import (
    find_font_file,
    get_default_system_font,
    get_image_font,
    get_plot_font,
    load_image,
    load_images,
    scale_image,
)

__all__ = [
    "GRAPHICs",
    "collage",
    "Collage",
    "gallery",
    "make_gif",
    "create_video",
    "extract_frames",
    "load_image",
    "load_images",
    "scale_image",
    "get_image_font",
    "get_default_system_font",
    "get_plot_font",
    "find_font_file",
]



"""
    HyFI Core Module
"""
import os
from pathlib import Path
from typing import List, Optional, Set, Tuple, Union

from pydantic import BaseModel

from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)

__hydra_version_base__: str = "1.2"
__hydra_default_config_group_value__: str = "__init__"
__hyfi_name__: str = "HyFI"
__hyfi_config_dirname__: str = "conf"
__hyfi_config_name__: str = "config"
__hyfi_user_config_path__: str = "config"
__hyfi_package_path__: str = Path(__file__).parent.parent.as_posix()
__hyfi_package_name__: str = os.path.basename(__hyfi_package_path__)
__hyfi_config_module__: str = f"{__hyfi_package_name__}.{__hyfi_config_dirname__}"
__hyfi_config_module_path__: str = f"pkg://{__hyfi_config_module__}"
__global_package_list__: Set[str] = {"cmd", "mode", "workflow"}

_batcher_instance_ = None


def __hyfi_version__() -> str:
    """
    Returns the version of HyFI.

    Returns:
        string containing the version of HyFI
    """
    from hyfi._version import __version__

    return __version__


class GlobalHyFIConfig(BaseModel):
    """Global configuration for HyFI

    Attributes:
    __package_name__ (str): The name of the package.
    __package_path__ (str): The path to the package root folder.
    __version__ (str): The version number of the package.
    __plugins__ (List[Any]): A list of plugins to load.
    __config_name__ (str): The name of the configuration module.
    __config_dirname__ (str): The name of the configuration directory.
    __user_config_path__ (str): The path to the user configuration directory.
    __dotenv_file__ (str): The name of the dotenv file.
    __secrets_dir__ (str): The name of the secrets directory.
    """

    __package_name__: str = __hyfi_package_name__
    __package_path__: str = __hyfi_package_path__
    __version__: str = __hyfi_version__()
    __plugins__: Optional[List[str]] = None

    __config_name__: str = __hyfi_config_name__
    __config_dirname__: str = __hyfi_config_dirname__
    __user_config_path__: str = __hyfi_user_config_path__

    __dotenv_file__: str = ".env"
    __secrets_dir__: str = "./secrets"

    __verbosity__: Union[bool, int] = False

    _packages_: List[Tuple[str, str]] = [(__hyfi_package_path__, __hyfi_version__())]

    def initialize(
        self,
        package_path: str = __hyfi_name__,
        version: str = __hyfi_version__(),
        plugins: Optional[List[str]] = None,
        user_config_path: Optional[str] = None,
        config_dirname: Optional[str] = None,
        dotenv_file: Optional[str] = None,
        secrets_dir: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Initializes the global HyFI instance.

        This function should be called before any other HyFI function.

        A plugin is a python module which contains a configuration module.

        Be careful!
        It does not check if the plugin is importable.

        Args:
            package_path: Path to the package root folder. e.g. `./src/hyfi`
            version: Version of the package. e.g. `0.1.0`
            plugins: A list of plugins to load. e.g. `["hyfi.conf"]`
            user_config_path: Path to the user configuration directory. e.g. `./config`
            config_dirname: Name of the configuration directory. e.g. `conf`
            dotenv_file: Name of the dotenv file. e.g. `.env`
            secrets_dir: Name of the secrets directory. e.g. `secrets`
            **kwargs: Additional arguments to be set as attributes.
        """
        self.__package_path__ = package_path
        self.__version__ = version
        if package_path not in self._packages_:
            self._packages_.append((package_path, version))
        if plugins:
            self.__plugins__ = self.get_plugins(plugins)
        if user_config_path:
            self.__user_config_path__ = user_config_path
        if config_dirname:
            self.__config_dirname__ = config_dirname
        if dotenv_file:
            self.__dotenv_file__ = dotenv_file
        if secrets_dir:
            self.__secrets_dir__ = secrets_dir

        # for the future use
        for key, value in kwargs.items():
            key_name = (
                key if key.startswith("__") and key.endswith("__") else f"__{key}__"
            )
            if hasattr(self, key_name):
                setattr(self, key, value)
            else:
                logger.warning("Invalid key: %s", key)

    @property
    def dotenv_file(self) -> str:
        """Returns the name of the dotenv file."""
        if os.environ.get("DOTENV_FILE"):
            return os.environ.get("DOTENV_FILE")
        return self.__dotenv_file__

    @property
    def secrets_dir(self) -> str:
        """Returns the path of the secrets directory."""
        secret_dir = self.__secrets_dir__
        if not os.path.isdir(secret_dir):
            secret_dir = os.environ.get("HYFI_SECRETS_DIR", "")
        if os.path.isdir(secret_dir):
            secret_dir = (
                secret_dir
                if os.path.isabs(secret_dir)
                else os.path.join(os.getcwd(), secret_dir)
            )
        else:
            logger.debug("The secrets directory does not exist.")
            secret_dir = None
        return secret_dir

    @property
    def plugins(self) -> Optional[List[str]]:
        """Returns the list of plugins to load."""
        return self.__plugins__

    def get_plugins(self, plugins: List[str]) -> List[str]:
        """Returns the list of plugins to load.
        A plugin is a python module which contains a configuration module.

        Be careful!
        It does not check if the plugin is importable.

        Args:
            plugins: List[str]: A list of plugins to load.
        """
        _plugins = []
        for plugin in plugins:
            plugin = plugin.split(".")[0]
            config_module = f"{plugin}.{self.__config_dirname__}"
            _plugins.append(config_module)
        return _plugins

    @property
    def package_name(self) -> str:
        """Returns the name of the package."""
        self.__package_name__ = os.path.basename(self.package_path)
        return self.__package_name__

    @property
    def package_path(self) -> str:
        """Returns the path to the package root folder.

        If there are multiple packages, the second package is returned.
        If there is only one package, the first package is returned. (default: hyfi)
        """
        if len(self._packages_) > 1:
            self.__package_path__ = self._packages_[1][0]
        else:
            self.__package_path__ = self._packages_[0][0]
        return self.__package_path__

    @property
    def version(self) -> str:
        """Returns the version number of the package."""
        if len(self._packages_) > 1:
            self.__version__ = self._packages_[1][1]
        else:
            self.__version__ = self._packages_[0][1]
        return self.__version__

    @property
    def config_dirname(self) -> str:
        """Returns the name of the configuration directory."""
        return self.__config_dirname__

    @property
    def config_root(self) -> str:
        """Returns the path to the configuration root directory."""
        return f"{self.package_path}/{self.config_dirname}"

    @property
    def config_module(self) -> str:
        """Returns the name of the configuration module."""
        return f"{self.package_name}.{self.config_dirname}"

    @property
    def config_module_path(self) -> str:
        """Returns the path to the configuration module."""
        return f"pkg://{self.config_module}"

    @property
    def config_name(self) -> str:
        """Returns the name of the configuration module."""
        return self.__config_name__

    @property
    def user_config_path(self) -> str:
        """Returns the path to the user configuration directory."""
        # if user_config_path is not an absolute path, make it absolute
        search_path = self.__user_config_path__
        if not os.path.isdir(search_path):
            search_path = os.environ.get("HYFI_USER_CONFIG_PATH", "")
        if os.path.isdir(search_path):
            self.__user_config_path__ = (
                search_path
                if os.path.isabs(search_path)
                else os.path.join(os.getcwd(), search_path)
            )
        else:
            logger.debug(
                "The user configuration directory does not exist: %s", search_path
            )
            self.__user_config_path__ = ""
        return self.__user_config_path__

    @property
    def hyfi_name(self) -> str:
        """Returns the name of HyFI package."""
        return __hyfi_name__

    @property
    def hyfi_config_module_path(self) -> str:
        """Returns the path to HyFI root folder"""
        return __hyfi_config_module_path__

    @property
    def hyfi_config_module(self) -> str:
        """Returns the name of HyFI default configuration module."""
        return __hyfi_config_module__

    @property
    def hyfi_config_name(self) -> str:
        """Returns the name of HyFI default configuration module."""
        return __hyfi_config_name__

    @property
    def hyfi_config_dirname(self) -> str:
        """Returns the name of HyFI default configuration directory."""
        return __hyfi_config_dirname__

    @property
    def hyfi_package_name(self) -> str:
        """Returns the name of HyFI package."""
        return __hyfi_package_name__

    @property
    def hyfi_package_path(self) -> str:
        """Returns the path to the package root folder."""
        return __hyfi_package_path__

    @property
    def hydra_version_base(self) -> str:
        """Returns the version of Hydra."""
        return __hydra_version_base__

    @property
    def hydra_default_config_group_value(self) -> str:
        """Returns the default config group value of Hydra."""
        return __hydra_default_config_group_value__

    @property
    def global_package_list(self) -> Set[str]:
        """Returns the list of global packages."""
        return __global_package_list__

    @property
    def verbosity(self) -> int:
        """Returns the verbosity level."""
        if os.environ.get("HYFI_VERBOSITY"):
            self.__verbosity__ = int(os.environ.get("HYFI_VERBOSITY", 0))
        return self.__verbosity__


__global_hyfi__ = GlobalHyFIConfig()


class GlobalHyFIResolver:
    @staticmethod
    def __hyfi_version__() -> str:
        """
        Returns the version of HyFI.

        Returns:
            string containing the version of HyFI
        """
        from hyfi._version import __version__

        return __version__

    def __hyfi_path__() -> str:
        """Returns the path to the HyFI root folder"""
        return __global_hyfi__.hyfi_package_path

    @staticmethod
    def __home_path__() -> str:
        """Returns the path to the user's home folder"""
        return Path.home().as_posix()

    @staticmethod
    def __app_version__() -> str:
        """
        Returns the version of App.

        Returns:
            string containing the version of App
        """

        return __global_hyfi__.version

    @staticmethod
    def __package_name__() -> str:
        """
        Returns the package name of the App

        Returns:
            string containing the package name of the App
        """

        return __global_hyfi__.package_name

    @staticmethod
    def __package_path__() -> str:
        """
        Returns the path to the App root folder

        Returns:
            string containing the path to the App root folder
        """

        return __global_hyfi__.package_path

    @staticmethod
    def __config_module_path__() -> str:
        """Global HyFI config path for the package to search for."""
        return __global_hyfi__.config_module_path

    @staticmethod
    def __user_config_path__() -> str:
        """Global HyFI user config path for the package to search for."""
        return __global_hyfi__.user_config_path



from typing import Dict, List, Optional, Union

from hyfi.pipeline.config import RunningCalls, RunningConfig, get_running_configs
from hyfi.run import RunConfig
from hyfi.task import TaskConfig
from hyfi.utils.contexts import elapsed_timer
from hyfi.utils.logging import LOGGING

logger = LOGGING.getLogger(__name__)


class BaseRunner(TaskConfig):
    _config_group_: str = "/runner"
    _config_name_: str = "__init__"

    calls: Optional[List[Union[str, Dict]]] = []

    def __call__(self) -> None:
        self.run()

    def run(self):
        self.run_calls()

    def get_running_calls(self) -> RunningCalls:
        return get_running_configs(self.calls or [])

    def run_call(self, rc: RunningConfig) -> None:
        method_ = getattr(self, rc.uses, None)
        if method_ and callable(method_):
            method_(**rc.run_kwargs)

    def run_calls(self):
        """
        Run the calls specified in the runner
        """
        calls = self.get_running_calls()
        if self.verbose:
            logger.info("Running %s call(s)", len(calls))
        # Run all tasks in the workflow.
        with elapsed_timer(format_time=True) as elapsed:
            for rc in calls:
                logger.info("Running call [%s] with [%s]", rc.uses, rc.run_kwargs)
                self.run_call(rc)
            if self.verbose:
                logger.info(
                    " >> elapsed time for the workflow with %s tasks: %s",
                    len(self.tasks or []),
                    elapsed(),
                )


class TestRunner(BaseRunner):
    _config_name_: str = "__test__"
    load_args: RunConfig = RunConfig(_config_name_="load_data")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def hello(self, **kwargs):
        print("Hello World!", kwargs)

    def world(self, **kwargs):
        print("World Hello!", kwargs)



"""
Configuration class for batch tasks. Inherits from TaskConfig.
"""
from pathlib import Path
from typing import Dict, List, Optional, Set, Union

from hyfi.batch import BatchConfig
from hyfi.composer import Composer
from hyfi.path.batch import BatchPathConfig
from hyfi.utils.logging import LOGGING

from .task import TaskConfig

logger = LOGGING.getLogger(__name__)


class BatchTaskConfig(TaskConfig):
    """
    Configuration class for batch tasks. Inherits from TaskConfig.

    Attributes:
        _config_name_ (str): The name of the configuration.
        _config_group_ (str): The configuration group.
        batch_name (str): The name of the batch.
        batch (BatchConfig): The batch configuration.
        _property_set_methods_ (Dict[str, str]): A dictionary of property set methods.
    """

    _config_name_: str = "__batch__"

    batch_name: str = "demo"
    batch: BatchConfig = BatchConfig()
    path: BatchPathConfig = BatchPathConfig()

    _property_set_methods_ = {
        "task_name": "set_task_name",
        "task_root": "set_task_root",
        "batch_name": "set_batch_name",
    }
    # _subconfigs_ = {"batch": BatchConfig}

    def set_batch_name(self, val: str):
        if not val:
            raise ValueError("Batch name cannot be empty")
        if not self.batch_name or self.batch_name != val:
            self.path.batch_name = val
            self.batch.batch_name = val

    def set_batch_num(self, val: Optional[int] = None):
        self.batch.batch_num = val

    def set_task_name(self, val: str):
        if not val:
            raise ValueError("Task name cannot be empty")
        if not self.task_name or self.task_name != val:
            self.path.task_name = val
            self.batch.batch_root = str(self.output_dir)

    def set_task_root(self, val: Union[str, Path]):
        if not val:
            raise ValueError("Task root cannot be empty")
        if not self.task_root or self.task_root != val:
            self.path.task_root = str(val)
            self.batch.batch_root = str(self.path.task_dir)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # At init, set the batch root to the task root
        # Othertimes, the batch root is set to the task root when the task root is set
        self.batch.batch_root = str(self.path.task_dir)
        # Call set_batch_num to set the batch_num and batch_id correctly after init
        if self.batch.batch_num_auto:
            self.batch.set_batch_num()
        logger.info(
            "Initalized batch: %s(%s) in %s",
            self.batch_name,
            self.batch_num,
            self.batch_dir,
        )

    @property
    def batch_num(self) -> int:
        return self.batch.batch_num

    @batch_num.setter
    def batch_num(self, val: Optional[int]):
        self.set_batch_num(val)

    @property
    def batch_id(self) -> str:
        return self.batch.batch_id

    @property
    def seed(self) -> int:
        return self.batch.seed

    @property
    def batch_dir(self) -> Path:
        return self.batch.batch_dir

    @property
    def device(self) -> str:
        return self.batch.device

    @property
    def num_devices(self) -> int:
        return self.batch.num_devices

    def save_config(
        self,
        filepath: Optional[Union[str, Path]] = None,
        exclude: Optional[Union[str, List[str], Set[str], None]] = None,
        exclude_none: bool = True,
        only_include: Optional[Union[str, List[str], Set[str], None]] = None,
        save_as_json_as_well: bool = True,
    ) -> str:
        """
        Save the batch configuration to file.

        Args:
            filepath (Optional[Union[str, Path]]): The filepath to save the configuration to. Defaults to None.
            exclude (Optional[Union[str, List[str], Set[str], None]]): Keys to exclude from the saved configuration.
                Defaults to None.
            exclude_none (bool): Whether to exclude keys with None values from the saved configuration. Defaults to True.
            only_include (Optional[Union[str, List[str], Set[str], None]]): Keys to include in the saved configuration.
                Defaults to None.
            save_as_json_as_well (bool): Whether to save the configuration as a json file as well. Defaults to True.

        Returns:
            str: The filename of the saved configuration.
        """
        if not self.batch:
            raise ValueError("No batch configuration to save")
        if not filepath:
            filepath = self.batch.config_filepath

        if save_as_json_as_well:
            self.save_config_as_json(
                exclude=exclude,
                exclude_none=exclude_none,
                only_include=only_include,
            )
        return super().save_config(
            filepath=filepath,
            exclude=exclude,
            exclude_none=exclude_none,
            only_include=only_include,
        )

    def save_config_as_json(
        self,
        filepath: Optional[Union[str, Path]] = None,
        exclude: Optional[Union[str, List[str], Set[str], None]] = None,
        exclude_none: bool = True,
        only_include: Optional[Union[str, List[str], Set[str], None]] = None,
    ) -> str:
        if not self.batch:
            raise ValueError("No batch configuration to save")
        if not filepath:
            filepath = self.batch.config_jsonpath
        return super().save_config_as_json(
            filepath=filepath,
            exclude=exclude,
            exclude_none=exclude_none,
            only_include=only_include,
        )

    def load_config(
        self,
        batch_name: Optional[str] = None,
        batch_num: Optional[int] = None,
        filepath: Optional[Union[str, Path]] = None,
        **config_kwargs,
    ) -> Dict:
        """Load the config from the batch config file"""
        if not self.batch:
            raise ValueError("No batch configuration to load")
        if not batch_name:
            batch_name = self.batch_name
        if batch_num is None:
            batch_num = -1
        if not filepath and batch_num >= 0:
            batch = BatchConfig(
                batch_root=self.batch.batch_root,
                batch_name=batch_name,
                batch_num=batch_num,
            )
            filepath = batch.config_filepath
        if isinstance(filepath, str):
            filepath = Path(filepath)

        if self.verbose:
            logger.info(
                "> Loading config for batch_name: %s batch_num: %s",
                batch_name,
                batch_num,
            )
        cfg = self.export_config()
        if filepath:
            if filepath.is_file():
                logger.info("Loading config from %s", filepath)
                batch_cfg = Composer.load(filepath)
                logger.info("Merging config with the loaded config")
                cfg = Composer.merge(cfg, batch_cfg)
            else:
                logger.info("No config file found at %s", filepath)
        if self.verbose:
            logger.info("Updating config with config_kwargs: %s", config_kwargs)
        cfg = Composer.update(Composer.to_dict(cfg), config_kwargs)

        # initialize self with the config
        self.__init__(**cfg)

        return self.model_dump()

    def print_config(
        self,
        batch_name: Optional[str] = None,
        batch_num: Optional[int] = None,
    ):
        self.load_config(batch_name, batch_num)
        Composer.print(self.model_dump())



from .aggregate import DSAggregate
from .basic import DSBasic
from .combine import DSCombine
from .load import DSLoad
from .plot import DSPlot
from .reshape import DSReshape
from .save import DSSave
from .slice import DSSlice
from .types import DatasetDictType, DatasetLikeType, DatasetType
from .utils import DSUtils


class DATASETs(
    DSAggregate,
    DSBasic,
    DSCombine,
    DSLoad,
    DSPlot,
    DSReshape,
    DSSave,
    DSSlice,
    DSUtils,
):
    def __init__(self):
        super().__init__()


__all__ = [
    "DatasetDictType",
    "DatasetLikeType",
    "DATASETs",
    "DatasetType",
    "DSLoad",
    "DSUtils",
]
